"""Class of LoggingRequestsWebClientInterceptor."""
import logging
from http.client import HTTPConnection
from typing import List

from .. import RequestsWebClient, WebClient, RequestDetails, RequestResponse


class LoggingRequestsWebClientInterceptor(WebClient.WebClientInterceptor):
    """Class of LoggingRequestsWebClientInterceptor."""

    @staticmethod
    def _debug_requests_on():
        """Switches on logging of the requests module."""
        HTTPConnection.debuglevel = 1

        logging.basicConfig()
        logging.getLogger().setLevel(logging.DEBUG)
        requests_log = logging.getLogger("requests.packages.urllib3")
        requests_log.setLevel(logging.DEBUG)
        requests_log.propagate = True

    @staticmethod
    def _debug_requests_off():
        """Switches off logging of the requests module, might be some side-effects."""
        HTTPConnection.debuglevel = 0

        root_logger = logging.getLogger()
        root_logger.setLevel(logging.WARNING)
        root_logger.handlers = []
        requests_log = logging.getLogger("requests.packages.urllib3")
        requests_log.setLevel(logging.NOTSET)
        requests_log.propagate = False

    def logs_to_show(self, params: RequestDetails) -> bool:
        """Method to decide that show logs of request.

        Method can be overridden and then the logs will be filtered – example by request url.
        """
        return True

    def intercept(
            self,
            requests_details: RequestDetails,
            next_interceptors: List[WebClient.WebClientInterceptor],
            web_client: RequestsWebClient
    ) -> RequestResponse:
        """Method show logs when predicate is true. Uses static field so it can be problem with concurrency."""
        is_to_log = self.logs_to_show(requests_details)
        if is_to_log:
            LoggingRequestsWebClientInterceptor._debug_requests_on()
        to_return = self.get_response(requests_details, next_interceptors, web_client)
        if is_to_log:
            LoggingRequestsWebClientInterceptor._debug_requests_off()
        return to_return



"""Class of ParamsResponseLogWebClientInterceptor."""
import threading
from typing import List

from .. import RequestsWebClient, WebClient, RequestDetails, RequestResponse


class ParamsResponseLogWebClientInterceptor(WebClient.WebClientInterceptor):
    """Class of ParamsResponseLogWebClientInterceptor.

    Interceptor log input params and out response.
    """

    _counter: int
    _lock: threading.Lock

    def __init__(self):
        """Constructor of ParamsResponseLogWebClientInterceptor."""
        self._value = 0
        self._lock = threading.Lock()

    def increment(self) -> int:
        """Thread safe increment. Returns old value."""
        with self._lock:
            to_return = self._value
            self._value += 1
            return to_return

    def logs_to_show(self, params: RequestDetails) -> bool:
        """Method to decide that show logs of request.

        Method can be overridden and then the logs will be filtered – example by request url.
        """
        return True

    def intercept(
            self,
            requests_details: RequestDetails,
            next_interceptors: List[WebClient.WebClientInterceptor],
            web_client: RequestsWebClient
    ) -> RequestResponse:
        """Method show logs when predicate is true. Uses static field so it can be problem with concurrency."""
        is_to_log = self.logs_to_show(requests_details)
        index = self.increment()
        if is_to_log:
            print(f'{index} -- {requests_details}')
        to_return = self.get_response(requests_details, next_interceptors, web_client)
        if is_to_log:
            print(f'{index} -- {to_return}')
        return to_return



"""Web client abstract class."""
from __future__ import annotations

from abc import abstractmethod, ABC
from typing import List, Optional

from .request_details import RequestDetails
from .request_response import RequestResponse


def _run_request_with_interceptors(
        requests_details: RequestDetails,
        next_interceptors: List[WebClient.WebClientInterceptor],
        web_client: WebClient
) -> RequestResponse:
    return next_interceptors[0].intercept(requests_details, next_interceptors[1:], web_client) if len(
        next_interceptors) > 0 else web_client.run_clear_request(requests_details)


class WebClient:
    """Web client abstract class."""

    _interceptors: List[WebClientInterceptor]

    def __init__(self, interceptors: Optional[List[WebClientInterceptor]]):
        """Base constructor of class."""
        self._interceptors = [] if interceptors is None else interceptors

    def run_request(self, requests_details: RequestDetails) -> RequestResponse:
        """Method process the request. Method wrap request with interceptors."""
        return _run_request_with_interceptors(requests_details, self._interceptors, self)

    @abstractmethod
    def run_clear_request(self, params: RequestDetails) -> RequestResponse:
        """Abstract method to run only the request."""

    class WebClientInterceptor(ABC):
        """Abstract class of web client interceptor."""

        @staticmethod
        def get_response(
                requests_details: RequestDetails,
                next_interceptors: List[WebClient.WebClientInterceptor],
                web_client: WebClient
        ) -> RequestResponse:
            """Method process request. If any interceptor passes method wrap request with this."""
            return _run_request_with_interceptors(requests_details, next_interceptors, web_client)

        @abstractmethod
        def intercept(
                self,
                requests_details: RequestDetails,
                next_interceptors: List[WebClient.WebClientInterceptor],
                web_client: WebClient
        ) -> RequestResponse:
            """Interceptor method of request.

            Method need to call WebClientInterceptor.get_response to process request by next interceptors
            and client.
            """



"""Mapping util."""
import json
from typing import List

from ..model.media import Media

_list_separator = ' , '


def simple_string_list_to_string(values: List[str]) -> str:
    """Method to build string from list."""
    return _list_separator.join(values)


def media_list_to_string(values: List[Media]) -> str:
    """Method to build string from Media list."""
    return json.dumps([it.__dict__() for it in values])


def string_to_simple_string_list(value: str) -> List[str]:
    """Method to build list from specific string."""
    return value.split(_list_separator) if len(value) > 0 else []


def string_to_media_list(value: str) -> List[Media]:
    """Method to build Media list from specific string."""
    parsed_json = json.loads(value)
    return [Media.from_dict(it) for it in parsed_json]



"""Domain Media class used to store media in Tweet."""
from dataclasses import dataclass
from typing import Dict

_URL = 'url'
_TYPE = 'type'


@dataclass
class Media:
    """Domain Media class used to store media in Tweet."""

    url: str
    type: str

    @classmethod
    def from_dict(cls, dictionary: Dict[str, str]):
        """Method to create Media class from dict."""
        return cls(dictionary[_URL], dictionary[_TYPE])

    def __dict__(self) -> Dict[str, str]:
        """Convert Media object to dictionary."""
        return dict({
            _URL: self.url,
            _TYPE: self.type
        })



"""Tweet - json mapper."""
import json

from arrow import Arrow

from .tweet_dict_mapper import create_tweet_from_dict
from ..model import Tweet
from ..model.media import Media


def tweet_to_json(tweet: Tweet) -> str:
    """Method to prepare json of tweet. Used in JSON serialization."""
    return json.dumps(tweet, cls=TweetJsonEncoder)


def create_tweet_from_json(json_value: str) -> Tweet:
    """Method creates tweet from json string."""
    tweet_dict = json.loads(json_value)
    if 'media' not in tweet_dict:
        tweet_dict['media'] = []
    else:
        tweet_dict['media'] = [Media.from_dict(it) for it in tweet_dict['media']]
    return create_tweet_from_dict(tweet_dict)


class TweetJsonEncoder(json.JSONEncoder):
    """JSONEncoder of tweet object."""

    def default(self, obj):
        """Method to convert object to JSON."""
        if isinstance(obj, Arrow):
            return str(obj)
        elif isinstance(obj, Media):
            return obj.__dict__()
        else:
            return obj.__dict__
